from datetime import datetime
import logging
import re

from multicorn import ForeignDataWrapper, ANY, ALL
from multicorn.utils import log_to_postgres

import elasticsearch
from elasticsearch.helpers import scan

from .es_helper import MatchList, get_filtered_query


class ESForeignDataWrapper(ForeignDataWrapper):

    # These operators and negated variants (prefixed with a '!') can be
    # offloaded to ElasticSearch.
    _PUSHED_DOWN_OPERATORS = ['=', '<>', '~~', '<@', '<', '>', '<=', '>=']

    # How many results should be returned in response to a single scroll request.
    # A lower value increases the number of round trips to the ElasticSearch
    # server to fetch the data.
    # A higher value increases the memory pressure on the ElasticSearch client
    # node issuing the search.
    _SCROLL_SIZE = 5000

    # How long a consistent view of the index should be maintained for scrolled
    # search. The default is 5 minutes.
    # This should be greater than the time it takes for an ES query to complete.
    # On the other hand, unnecessarily high values needlessly increase the load
    # on the cluster.
    _SCROLL_LENGTH = '5m'

    def __init__(self, options, columns):
        super(ESForeignDataWrapper, self).__init__(options, columns)
        self._esclient = None
        self._options = options
        self._columns = columns
        self._doc_type = options['doc_type']
        if options.get('column_name_translation') == 'true':
            self._column_to_es_field = self.convert_column_name
        else:
            self._column_to_es_field = lambda column: column

    @property
    def esclient(self):
        if self._esclient is None:
            params = {}
            if 'hostname' in self._options:
                params['host'] = self._options['hostname']
            if 'port' in self._options:
                params['port'] = self._options['port']
            self._esclient = elasticsearch.Elasticsearch([params])
        return self._esclient

    def get_index(self, _quals):
        """Get the ElasticSearch index or indices to query.

        By default, we obtain the index from the foreign table options.
        However, this method can be overridden to derive the index from the query
        quals. For example, the `timestamp` qual could be used to select one or
        more time-based indices.
        """
        return self._options['index']

    def convert_column_name(self, column):
        """Given a column name, return the corresponding Elasticsearch field name.

        The default implementation replaces `__` with `.` (allowing for specifying
        fields in nested objects) and `_` with `-` (to reflect common ElasticSearch
        convention).

        `_id` is left untranslated.
        `timestamp` is converted to `@timestamp` to match Logstash conventions.

        This method is used only for foreign tables that were created with the option
        `column_name_translation` set to the value `true`.

        This method can be overridden in a subclass if a different implementation
        is desired.
        """
        if column == '_id':
            return column
        elif column == 'timestamp':
            return '@timestamp'
        return column.replace('__', '.').replace('_', '-')

    def _endpoint_to_datetime(self, endpoint):
        # When dealing with date and time ranges, we get a string formatted as
        # %Y-%m-%d %H:%M:%S(.%f)?
        valid_time_formats = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f']
        for fmt in valid_time_formats:
            try:
                return datetime.strptime(endpoint, fmt)
            except ValueError:
                pass
        # No expected time formats matched, so no conversion
        return endpoint

    def _append_filter(self, filter_list, field, operator, value):
        if operator == '=':
            if value is None:
                filter_list.append_missing(field)
            else:
                filter_list.append_term(field, value)
        elif operator == '<>':
            if value is None:
                filter_list.append_exists(field)
            else:
                filter_list.append_term(field, value, not_value=True)
        elif operator == '~~':
            # LIKE operator
            if value.find('%') == len(value) - 1 and '_' not in value:
                filter_list.append_prefix(field, value[:-1])
            else:
                value = re.escape(value).replace(
                    r'\%', '.*').replace(r'\_', '.')
                filter_list.append_filter({'regexp': {field: value}})
        elif operator == '<@':
            # range operator
            # value looks something like ["start","end")
            # (including quotation marks when dealing with date strings)
            def _format_endpoint(endpoint):
                endpoint = self._endpoint_to_datetime(endpoint)
                if isinstance(endpoint, datetime):
                    return endpoint.strftime('%Y-%m-%dT%H:%M:%S.%f')
                return endpoint

            start, end = value.split(',')
            if start[0] == '[':
                start_op = 'gte'
            else:
                start_op = 'gt'
            start = start[1:].replace('"', '')

            if end[-1] == ']':
                end_op = 'lte'
            else:
                end_op = 'lt'
            end = end[:-1].replace('"', '')

            params = {
                start_op: _format_endpoint(start),
                end_op: _format_endpoint(end)
            }
            filter_list.append_range(field, **params)
        elif operator in ('<', '<=', '>', '>='):
            operator_map = {
                '<': 'lt',
                '<=': 'lte',
                '>': 'gt',
                '>=': 'gte'
            }
            params = {
                operator_map[operator]: value
            }
            filter_list.append_range(field, **params)

    def _normalize_operator(self, operator, value):
        negation = False
        if operator[0] == '!':
            # Negative operators are handled as their positive counterparts
            # but are added to the must_not list.
            operator = operator[1:]
            negation = True
        if operator == '<>' and value is not None:
            # Generally handle a not-equals in the must_not list.
            # IS NOT NULL, which is passed down to us as '<> None', is handled
            # as a `missing` filter in the must_list.
            operator = '='
            negation = True
        return operator, negation

    def _process_qual(self, must_list, must_not_list, field, operator, value):
        operator, negation = self._normalize_operator(operator, value)
        if operator in self._PUSHED_DOWN_OPERATORS:
            filter_list = must_not_list if negation else must_list
            self._append_filter(filter_list, field, operator, value)

    def _make_match_lists(self, quals):
        must_list = MatchList()
        must_not_list = MatchList()
        for qual in quals:
            field = self._column_to_es_field(qual.field_name)
            if qual.list_any_or_all == ANY:
                if qual.operator[0] == '=':
                    must_list.append_terms(field, qual.value)
                else:
                    match_list = MatchList()
                    operator, negation = self._normalize_operator(
                        qual.operator[0], True)
                    for elem in qual.value:
                        self._append_filter(match_list, field, operator, elem)
                    if negation:
                        # a <> any(x, y, z) => a <> x or a <> y or a <> z =>
                        # not (a = x and a = y and a = z)
                        must_not_list.append_filter({'and': match_list})
                    else:
                        must_list.append_filter({'or': match_list})
            elif qual.list_any_or_all == ALL:
                if qual.operator[0] == '<>':
                    # a <> all(x, y, z) => a <> x and a <> y and a <> z => not
                    # (a = x or a = y or a = z)
                    must_not_list.append_terms(field, qual.value)
                else:
                    for elem in qual.value:
                        self._process_qual(
                            must_list, must_not_list, field, qual.operator[0], elem)
            else:
                self._process_qual(
                    must_list,
                    must_not_list,
                    field,
                    qual.operator,
                    qual.value)
        return must_list, must_not_list

    def execute(self, quals, columns, _sortkeys=None):
        must_list, must_not_list = self._make_match_lists(quals)
        if must_list or must_not_list:
            query = get_filtered_query(
                must_list=must_list,
                must_not_list=must_not_list)
        else:
            query = {}
        # It's not clear if we should be using `fields` or `_source` here.
        # `fields` is useful for "stored" fields, which are stored separately
        # from the main _source JSON document. The idea is that the entire document
        # does not need to be reparsed when loading only a subset of the fields.
        # When fields aren't "stored" but are doc_values, they still seem to be
        # stored independently.
        # Tests suggest that at least in some cases when dealing with doc_values,
        # `fields` 1.16 times better than `_source`.
        query['fields'] = [self._column_to_es_field(
            column) for column in columns]
        # When using fields, the values always come back in an array, to make for
        # more consistent treatment of any actual array fields that we may have
        # requested. If the field is not truly an array field, the value comes back
        # in an array of one element.
        default_value = [None]
        log_to_postgres('query: %s' % query, logging.DEBUG)
        for result in scan(
                self.esclient,
                query=query,
                index=self.get_index(quals),
                doc_type=self._doc_type,
                size=self._SCROLL_SIZE,
                scroll=self._SCROLL_LENGTH):
            obs = result.get('fields', {})

            def _massage_value(value, column):
                if column == '_id':
                    # `_id` is special in that it's always present in the top-level
                    # result, not under `fields`.
                    return result['_id']
                # If the column type is an array, return the list.
                # Otherwise, return the first element of the array.
                if self._columns[column].type_name.endswith('[]'):
                    return value
                return value[0]
            row = {
                column: _massage_value(
                    obs.get(
                        self._column_to_es_field(column),
                        default_value),
                    column) for column in columns}
            yield row

    def get_rel_size(self, quals, columns):
        must_list, must_not_list = self._make_match_lists(quals)
        if must_list or must_not_list:
            query = get_filtered_query(
                must_list=must_list,
                must_not_list=must_not_list)
        else:
            query = {}
        query['size'] = 0
        results = self.esclient.search(
            index=self.get_index(quals),
            body=query,
            doc_type=self._doc_type)
        return (results['hits']['total'], len(columns) * 100)



from __future__ import print_function
import argparse
from collections import namedtuple
import copy
import json
import sys


ColumnSpec = namedtuple('ColumnSpec', 'column_name data_type')
TableSpec = namedtuple('TableSpec', 'name columns doc_type index')
# Separator for name components of nested objects, used to construct column
# names for a nested field.
# i.e., foo.bar.baz becomes foo__bar__baz
NESTED_FIELD_SEPARATOR = '__'
# A map from JSON types to PostgreSQL types.
# Types not in the map are assumed to have the same name in both type systems.
TYPE_MAP = {
    'string': 'text',
    'long': 'bigint',
    'short': 'smallint',
    'double': 'double precision',
    'date': 'timestamp'
}


def translate_es_name(name):
    """Translate Elasticsearch name to PostgreSQL name.

    Replaces dashes with underscores.
    """
    if name == '@timestamp':
        return 'timestamp'
    return name.replace('-', '_')


def generate_columns(mapping, path=None):
    """Given a dict representing ES field mappings for a particular object,
    generate ColumnSpec objects representing the corresponding columns.

    Nested objects are handled recursively, with `generate_columns` called for
    the mapping of each object. The recursive calls include a path argument,
    a list of the field names of the outer objects in which the current
    object is nested. This path is used to construct the column name.
    """
    if path is None:
        path = []
    for field in mapping:
        path_to_field = copy.copy(path)
        path_to_field.append(translate_es_name(field))
        if 'type' in mapping[field]:
            es_type = mapping[field]['type']
            yield ColumnSpec(NESTED_FIELD_SEPARATOR.join(path_to_field),
                             TYPE_MAP.get(es_type, es_type))
        else:
            # Nested object
            for column_spec in generate_columns(
                    mapping[field]['properties'], path=path_to_field):
                yield column_spec


def generate_table_spec(mapping, include_indices, include_doc_types):
    """Generate TableSpec objects corresponding to the mappings in `mapping`,
    a dict whose structure is expected to match the JSON that is returned by the
    get mapping API (https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-get-mapping.html).

    One TableSpec is generated per doc_type per index, with the name of the
    table derived from the doc_type.

    `include_indices` is a list that contains the indices that should be processed.
    If empty or None, all indices in the mapping are processed.

    `include_doc_types` is a list that contains the doc_types for which tables
    should be generated.
    If empty or None, all doc_types are included.
    """
    for index in mapping:
        if include_indices and index not in include_indices:
            continue
        for doc_type in mapping[index]['mappings']:
            # _default_ and .percolator are possible mapping keys that are not
            # doc_types
            if doc_type in (
                    '_default_',
                    '.percolator') or include_doc_types and doc_type not in include_doc_types:
                continue
            table = translate_es_name(doc_type)
            columns = list(
                generate_columns(
                    mapping[index]['mappings'][doc_type]['properties']))
            if columns:
                yield TableSpec(table, columns, doc_type, index)


def generate_schema(mapping, include_indices, include_doc_types, server):
    """Generate `DROP FOREIGN TABLE` and `CREATE FOREIGN TABLE` DDL from
    Elasticsearch field mappings.

    `server` is the name of the entity created with `CREATE SERVER`
    to use in the foreign table schema.

    All other arguments are as described in `generate_table_spec`.
    """
    for table_spec in generate_table_spec(
            mapping, include_indices, include_doc_types):
        columns = ',\n'.join(
            '    %s %s' %
            (col.column_name, col.data_type) for col in table_spec.columns)
        yield \
            """DROP FOREIGN TABLE IF EXISTS %(table)s;
CREATE FOREIGN TABLE %(table)s (
%(columns)s
) SERVER %(server)s OPTIONS (
    doc_type '%(doc_type)s',
    index '%(index)s',
    column_name_translation 'true'
);
""" % {'table': table_spec.name, 'columns': columns, 'server': server,
                'doc_type': table_spec.doc_type, 'index': table_spec.index}


def main():
    description = \
        """Generate foreign table schema from Elasticsearch mappings

Given a JSON file with Elasticsearch mappings (see https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping.html
for the format and https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-get-mapping.html
for an explanation of how to extract from a running cluster), generate
`CREATE FOREIGN TABLE` statements to create tables for use with esfdw.

The script generates a foreign table per doc_type per index, with the name of
the table derived from the name of the doc_type, except with dashes replaced with
underscores.

Column types are translated from the Elasticsearch equivalent and are always
scalar. Nested objects are not represented as JSON; instead, a column definition
is generated for every nested leaf field. Elasticsearch mappings do not contain
an indication of whether the field is a list field, which means that the script
cannot know when to make a column an array. The schema can be fixed up manually
if array columns are desired.

Column names match the Elasticsearch field names except for the following:
  * Dashes are replaced with single underscores
  * The field separator for nested objects is `__` rather than `.`
    That is, the Elasticsearch field foo.bar.baz is mapped to the column name foo__bar__baz
"""
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        '-d',
        '--doc-types',
        nargs='+',
        help='doc_type names for which to generate foreign table definitions')
    parser.add_argument(
        '-i',
        '--indices',
        nargs='+',
        help='names of indices whose mappings should be processed when generating foreign table definitions')
    parser.add_argument(
        '-s',
        '--server',
        required=True,
        help='name of the PostgreSQL server object to use in the foreign table definition')
    args = parser.parse_args()

    mapping = json.load(sys.stdin)
    print(
        '\n'.join(
            generate_schema(
                mapping,
                args.indices,
                args.doc_types,
                args.server)))


if __name__ == "__main__":
    main()



class MatchList(list):
    """Helper for building ES query objects. Boolean queries take a list of sub-terms. This
    is designed to be able programmatically build the sub-terms. This object can be
    treated as a list.
    """

    def append_filter(self, new_filter, not_value=False, constant_score=None):
        """Append a filter to our list. Enclose in a 'not' if requested"""
        if not_value:
            new_filter = {
                'not': new_filter
            }
        if constant_score:
            new_filter = {
                'constant_score': {
                    'filter': new_filter,
                    'boost': constant_score
                }
            }
        self.append(new_filter)

    def append_range(
            self,
            key,
            not_value=False,
            constant_score=None,
            **kwargs):
        """Append a range filter.

        See http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-range-filter.html
        for parameters
        """
        self.append_filter(
            {
                'range': {key: kwargs}
            }, not_value, constant_score
        )

    def append_exists(self, key, not_value=False, constant_score=None):
        """An exists test. Add to a must-not filter to test for not exists"""
        self.append_filter(
            {
                'exists': {'field': key}
            }, not_value, constant_score
        )

    def append_missing(self, key, not_value=False, constant_score=None):
        """A missing test."""
        self.append_filter(
            {
                'missing': {'field': key}
            }, not_value, constant_score
        )

    def append_term(self, key, value, not_value=False, constant_score=None):
        """Single term filter. See if the given key has the given value"""
        if value is not None:
            self.append_filter(
                {
                    'term': {key: value}
                }, not_value, constant_score
            )

    def append_prefix(self, key, value, not_value=False, constant_score=None):
        """Single prefix filter. See if the given key has the given value as its prefix"""
        if value is not None:
            self.append_filter(
                {
                    'prefix': {key: value}
                }, not_value, constant_score
            )

    def append_terms(
            self,
            key,
            value_list,
            not_value=False,
            constant_score=None):
        """Multi-term filter. See if the given key has one of the given value_list"""
        if value_list is not None:
            self.append_filter(
                {
                    'terms': {key: value_list}
                }, not_value, constant_score
            )


def get_filtered_query(must_list=None, must_not_list=None):
    """Get the correct query string for a boolean filter. Accept must and
    must_not lists. Use MatchList for generating the appropriate lists.
    """
    bool_filter = {}
    if must_list:
        bool_filter['must'] = must_list
    if must_not_list:
        bool_filter['must_not'] = must_not_list
    result = {
        'query': {
            'filtered': {
                'filter': {
                    'bool': bool_filter
                }
            }
        }
    }
    return result



from setuptools import setup

long_description = \
"""esfdw is a PostgreSQL `foreign data
wrapper <http://www.postgresql.org/docs/current/static/postgres-fdw.html>`__
for Elasticsearch.

Elasticsearch is widely used for document and log data storage, in
particular as part of the `ELK
stack <https://www.elastic.co/webinars/introduction-elk-stack>`__. esfdw
allows PostgreSQL to be used as a query engine for data stored in
Elasticsearch. Use cases include:

- Writing SQL JOIN queries against Elasticsearch documents and letting the PostgreSQL engine do the heavy lifting
- Running window functions on data stored in Elasticsearch
- Applying PostgreSQL aggregations that do not currently have a native Elasticsearch equivalent

esfdw depends on `Multicorn <http://multicorn.org>`__, a PostgreSQL
extension for writing foreign data wrappers in Python.
"""

setup(
    name='esfdw',
    description='PostgreSQL foreign data wrapper for Elasticsearch',
    long_description=long_description,
    version='0.1.1',
    author='Arctic Wolf Networks, Inc.',
    author_email='info@arcticwolf.com',
    license='MIT',
    packages=['esfdw'],
    test_suite='tests',
    url='https://github.com/rtkwlf/esfdw',
    download_url='https://github.com/rtkwlf/esfdw/tarball/0.1.1'
)
